"""Functions to create full tensor networks backed by `symmray`."""

import numbers


def parse_edges_to_site_info(
    edges,
    bond_dim,
    phys_dim=2,
    site_ind_id="k{}",
    bond_ind_id="b{}-{}",
    site_tag_id="I{}",
):
    """Given a list of edges, return a dictionary of site information, each
    specifying the local shape, index identifiers, index dualnesses, and tags.
    The dualnesses of the bonds are set in a canonical order corresponding to
    sorting all the sites and the edges.

    Parameters
    ----------
    edges : Sequence[Tuple[hashable, hashable]]
        The edges of the graph.
    bond_dim : int
        The internal bond dimension.
    phys_dim : int, optional
        The physical dimension of the sites.
    site_ind_id : str, optional
        The identifier for the site indices.
    bond_ind_id : str, optional
        The identifier for the bond indices.
    site_tag_id : str, optional
        The identifier for the site tags.

    Returns
    -------
    Dict[hashable, Dict[str, Any]]
    """
    site_info = {}

    starmap_ind = site_ind_id.count("{}") > 1
    starmap_tag = site_tag_id.count("{}") > 1

    # create bonds
    for sitea, siteb in sorted(edges):
        if sitea > siteb:
            sitea, siteb = siteb, sitea

        ind = bond_ind_id.format(sitea, siteb)
        infoa = site_info.setdefault(sitea, {})
        infob = site_info.setdefault(siteb, {})

        infoa.setdefault("inds", []).append(ind)
        infob.setdefault("inds", []).append(ind)

        infoa.setdefault("duals", []).append(0)
        infob.setdefault("duals", []).append(1)

        infoa.setdefault("shape", []).append(bond_dim)
        infob.setdefault("shape", []).append(bond_dim)

    # create physical inds
    for site in site_info:
        site_info[site]["coordination"] = len(site_info[site]["inds"])

        if starmap_tag:
            site_tag = site_tag_id.format(*site)
        else:
            site_tag = site_tag_id.format(site)

        site_info[site]["tags"] = (site_tag,)

        if phys_dim is not None:
            if starmap_ind:
                site_ind = site_ind_id.format(*site)
            else:
                site_ind = site_ind_id.format(site)

            site_info[site]["inds"].append(site_ind)
            site_info[site]["duals"].append(0)
            site_info[site]["shape"].append(phys_dim)

    # put in canonical sorted by site order
    site_info = {k: site_info[k] for k in sorted(site_info)}

    return site_info


_DEFAULT_PHYS_CHARGEMAPS = {
    ("Z2", 2): {0: 1, 1: 1},
    ("U1", 2): {0: 1, 1: 1},
    ("Z2", 4): {0: 2, 1: 2},
    ("U1", 4): {0: 1, 1: 2, 2: 1},
    ("U1U1", 4): {(0, 0): 1, (0, 1): 1, (1, 0): 1, (1, 1): 1},
    ("Z2Z2", 4): {(0, 0): 1, (0, 1): 1, (1, 0): 1, (1, 1): 1},
}


def TN_abelian_from_edges_rand(
    symmetry,
    edges,
    bond_dim,
    phys_dim=None,
    seed=None,
    dtype="float64",
    site_tag_id="I{}",
    site_ind_id="k{}",
    fermionic=False,
    flat=False,
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Build a random abelian symmetric `quimb.tensor` amplitude or PEPS from
    edges.

    Parameters
    ----------
    symmetry : {"Z2", "U1", "Z2Z2", "U1U1", ZN}
        The symmetry of the PEPS. Currently only "Z2" and "U1" are supported.
    edges : list of tuples
        The edges of the PEPS. Each edge is a tuple of the form `(cooa, coob)`
        where cooa and coob are hashable labels of the two sites.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : None, int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, all sites have the same dimension. If a
        dictionary, a custom map of physical charges to sizes.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    fermionic : bool, optional
        Whether to generate fermionic tensors.
    flat : bool, optional
        Whether to generate 'flat' backend arrays (True) or the default
        block-sparse backend arrays (False).
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alternate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.TensorNetworkGen or quimb.tensor.TensorNetworkGenVector
    """
    import quimb.tensor as qtn

    import symmray as sr

    site_info = parse_edges_to_site_info(
        edges,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        site_ind_id=site_ind_id,
        site_tag_id=site_tag_id,
    )
    sites = tuple(site_info.keys())

    if phys_dim is not None:
        # have physical sites
        tn = qtn.tensor_arbgeom.TensorNetworkGenVector.new(
            sites=sites,
            site_ind_id=site_ind_id,
            site_tag_id=site_tag_id,
        )

        if isinstance(phys_dim, dict):
            # custom physical charge map
            phys_chargemap = phys_dim
        elif isinstance(phys_dim, numbers.Integral):
            phys_chargemap = _DEFAULT_PHYS_CHARGEMAPS[symmetry, phys_dim]
        else:
            phys_chargemap = phys_dim
    else:
        # no physical sites

        tn = qtn.tensor_arbgeom.TensorNetworkGen.new(
            sites=sites,
            site_tag_id=site_tag_id,
        )

        phys_chargemap = None

    if site_charge is None:
        from symmray.symmetries import ZN, get_symmetry

        if symmetry == "U1":
            even_sites = set(sites[::2])

            def site_charge(site):
                return 0 if site in even_sites else 1

        elif isinstance(get_symmetry(symmetry), ZN):

            def site_charge(site):
                return 0

        else:
            raise ValueError(
                f"symmetry={symmetry} has no default "
                "`site_charge`. Please provide one."
            )

    rng = sr.utils.get_rng(seed)

    index_store = {}
    for site, info in site_info.items():
        if phys_dim is None:
            shape = info["shape"]
        else:
            shape = info["shape"][:-1] + [phys_chargemap]
        duals = info["duals"]

        # resolve shape sizes into explicit indices ahead of time so that we
        # can build them in conj-pairs with matching subsizes even if random
        shape_parsed = []
        for ix, size, dual in zip(info["inds"], shape, duals):
            if ix in index_store:
                shape_parsed.append(index_store[ix].conj())
            else:
                index_store[ix] = sr.utils.rand_index(
                    symmetry,
                    size,
                    dual=dual,
                    subsizes=subsizes,
                    seed=rng,
                )
                shape_parsed.append(index_store[ix])

        tn |= qtn.Tensor(
            data=sr.utils.get_rand(
                shape=shape_parsed,
                duals=duals,
                symmetry=symmetry,
                charge=site_charge(site),
                fermionic=fermionic,
                flat=flat,
                subsizes=subsizes,
                seed=rng,
                dtype=dtype,
                # possibly needed for odd parity fermionic tensors
                label=site,
                **kwargs,
            ),
            inds=info["inds"],
            tags=info["tags"],
        )

    return tn


def TN_fermionic_from_edges_rand(
    symmetry,
    edges,
    bond_dim,
    phys_dim=None,
    seed=None,
    dtype="float64",
    site_tag_id="I{}",
    site_ind_id="k{}",
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random fermionic tensor network from edges. This is a wrapper
    around :func:`TN_abelian_from_edges_rand` with `fermionic=True`.

    Parameters
    ----------
    symmetry : {"Z2", "U1", "Z2Z2", "U1U1", ZN}
        The symmetry of the PEPS. Currently only "Z2" and "U1" are supported.
    edges : list of tuples
        The edges of the PEPS. Each edge is a tuple of the form `(cooa, coob)`
        where cooa and coob are hashable labels of the two sites.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : None, int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, all sites have the same dimension. If a
        dictionary, a custom map of physical charges to sizes.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.TensorNetworkGen or quimb.tensor.TensorNetworkGenVector
    """
    return TN_abelian_from_edges_rand(
        symmetry=symmetry,
        edges=edges,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        seed=seed,
        dtype=dtype,
        site_tag_id=site_tag_id,
        site_ind_id=site_ind_id,
        fermionic=True,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )


def MPS_abelian_rand(
    symmetry,
    L,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{}",
    site_ind_id="k{}",
    fermionic=False,
    flat=False,
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random MPS with abelian symmetry.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the MPS.
    L : int
        The number of sites.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the MPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the MPS is cyclic.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    fermionic : bool, optional
        Whether to generate fermionic tensors.
    flat : bool, optional
        Whether to generate 'flat' backend arrays (True) or the default
        block-sparse backend arrays (False).
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.MatrixProductState
    """
    import quimb.tensor as qtn

    edges = qtn.edges_1d_chain(L, cyclic=cyclic)

    mps = TN_abelian_from_edges_rand(
        symmetry,
        edges,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        seed=seed,
        dtype=dtype,
        site_ind_id=site_ind_id,
        site_tag_id=site_tag_id,
        fermionic=fermionic,
        flat=flat,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )

    return mps.view_as_(
        qtn.MatrixProductState,
        L=L,
        cyclic=cyclic,
    )


def PEPS_abelian_rand(
    symmetry,
    Lx,
    Ly,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{},{}",
    site_ind_id="k{},{}",
    x_tag_id="X{}",
    y_tag_id="Y{}",
    fermionic=False,
    flat=False,
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random 2D PEPS with abelian symmetry.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the PEPS.
    Lx : int
        The number of rows.
    Ly : int
        The number of columns.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the PEPS is cyclic in the x-direction.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    fermionic : bool, optional
        Whether to generate fermionic tensors.
    flat : bool, optional
        Whether to generate 'flat' backend arrays (True) or the default
        block-sparse backend arrays (False).
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.PEPS
    """
    import quimb.tensor as qtn

    edges = qtn.edges_2d_square(Lx, Ly, cyclic=cyclic)

    peps = TN_abelian_from_edges_rand(
        symmetry,
        edges,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        seed=seed,
        dtype=dtype,
        site_ind_id=site_ind_id,
        site_tag_id=site_tag_id,
        fermionic=fermionic,
        flat=flat,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )

    for i in range(Lx):
        for j in range(Ly):
            t = peps[site_tag_id.format(i, j)]
            t.add_tag(x_tag_id.format(i))
            t.add_tag(y_tag_id.format(j))

    return peps.view_as_(
        qtn.PEPS, Lx=Lx, Ly=Ly, x_tag_id=x_tag_id, y_tag_id=y_tag_id
    )


def PEPS3D_abelian_rand(
    symmetry,
    Lx,
    Ly,
    Lz,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{},{},{}",
    site_ind_id="k{},{},{}",
    x_tag_id="X{}",
    y_tag_id="Y{}",
    z_tag_id="Z{}",
    fermionic=False,
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random 3D PEPS with abelian symmetry.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the PEPS.
    Lx : int
        Length of the PEPS in the x-direction.
    Ly : int
        Length of the PEPS in the y-direction.
    Lz : int
        Length of the PEPS in the z-direction.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the PEPS is cyclic in the x-direction.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    fermionic : bool, optional
        Whether to generate fermionic tensors.
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.PEPS3D
    """
    import quimb.tensor as qtn

    edges = qtn.edges_3d_cubic(Lx, Ly, Lz, cyclic=cyclic)

    peps = TN_abelian_from_edges_rand(
        symmetry,
        edges,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        seed=seed,
        dtype=dtype,
        site_ind_id=site_ind_id,
        site_tag_id=site_tag_id,
        fermionic=fermionic,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )

    for i in range(Lx):
        for j in range(Ly):
            for k in range(Lz):
                t = peps[site_tag_id.format(i, j, k)]
                t.add_tag(x_tag_id.format(i))
                t.add_tag(y_tag_id.format(j))
                t.add_tag(z_tag_id.format(k))

    return peps.view_as_(
        qtn.PEPS3D,
        Lx=Lx,
        Ly=Ly,
        Lz=Lz,
        x_tag_id=x_tag_id,
        y_tag_id=y_tag_id,
        z_tag_id=z_tag_id,
    )


def MPS_fermionic_rand(
    symmetry,
    L,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{}",
    site_ind_id="k{}",
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random fermionic MPS. This is a wrapper around
    :func:`MPS_abelian_rand` with `fermionic=True`.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the MPS.
    L : int
        The number of sites.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the MPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the MPS is cyclic.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.MatrixProductState
    """
    return MPS_abelian_rand(
        symmetry=symmetry,
        L=L,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        cyclic=cyclic,
        seed=seed,
        dtype=dtype,
        site_tag_id=site_tag_id,
        site_ind_id=site_ind_id,
        fermionic=True,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )


def PEPS_fermionic_rand(
    symmetry,
    Lx,
    Ly,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{},{}",
    site_ind_id="k{},{}",
    x_tag_id="X{}",
    y_tag_id="Y{}",
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random 2D fermionic PEPS. This is a wrapper around
    :func:`PEPS_abelian_rand` with `fermionic=True`.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the PEPS.
    Lx : int
        The number of rows.
    Ly : int
        The number of columns.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the PEPS is cyclic in the x-direction.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.PEPS
    """
    return PEPS_abelian_rand(
        symmetry=symmetry,
        Lx=Lx,
        Ly=Ly,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        cyclic=cyclic,
        seed=seed,
        dtype=dtype,
        site_tag_id=site_tag_id,
        site_ind_id=site_ind_id,
        x_tag_id=x_tag_id,
        y_tag_id=y_tag_id,
        fermionic=True,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )


def PEPS3D_fermionic_rand(
    symmetry,
    Lx,
    Ly,
    Lz,
    bond_dim,
    phys_dim=2,
    cyclic=False,
    seed=None,
    dtype="float64",
    site_tag_id="I{},{},{}",
    site_ind_id="k{},{},{}",
    x_tag_id="X{}",
    y_tag_id="Y{}",
    z_tag_id="Z{}",
    site_charge=None,
    subsizes="maximal",
    **kwargs,
):
    """Create a random 3D fermionic PEPS. This is a wrapper around
    :func:`PEPS3D_abelian_rand` with `fermionic=True`.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the PEPS.
    Lx : int
        Length of the PEPS in the x-direction.
    Ly : int
        Length of the PEPS in the y-direction.
    Lz : int
        Length of the PEPS in the z-direction.
    bond_dim : int or dict
        The total (sum of charge sizes) bond dimension of the PEPS. You can
        also provide an explicit map of bond charges to sizes.
    phys_dim : int or dict, optional
        The physical dimension of each site. If None, no physical sites are
        included. If an integer, a default charge distribution is chosen. If a
        dictionary, a custom map of physical charges to sizes.
    cyclic : bool, optional
        Whether the PEPS is cyclic in the x-direction.
    seed : None, int or np.random.Generator, optional
        The random seed or generator to use.
    dtype : str, optional
        The data type of the tensors.
    site_tag_id : str, optional
        The tag format for each site tensor.
    site_ind_id : str, optional
        The index format for each site tensor, if physical sites are included.
    site_charge : callable, optional
        A function that takes a site index and returns the charge of that site.
        By default it will create all even parity tensors if Z2=0 or it will
        alernate between 0 and 1 for U1.
    subsizes : {"maximal", "equal"}, optional
        The sizes of the charge sectors. If None, the sizes are randomly
        determined. If "equal", the sizes are equal (up to remainders). If
        "maximal", as many charges as possible will be chosen.
    kwargs
        Additional arguments to pass to :func:`symmray.utils.get_rand`.

    Returns
    -------
    quimb.tensor.PEPS3D
    """
    return PEPS3D_abelian_rand(
        symmetry=symmetry,
        Lx=Lx,
        Ly=Ly,
        Lz=Lz,
        bond_dim=bond_dim,
        phys_dim=phys_dim,
        cyclic=cyclic,
        seed=seed,
        dtype=dtype,
        site_tag_id=site_tag_id,
        site_ind_id=site_ind_id,
        x_tag_id=x_tag_id,
        y_tag_id=y_tag_id,
        z_tag_id=z_tag_id,
        fermionic=True,
        site_charge=site_charge,
        subsizes=subsizes,
        **kwargs,
    )



"""Fermionic symmetric arrays with block sparse backend."""

import autoray as ar

from ..abelian_common import AbelianCommon
from ..common import SymmrayCommon
from ..fermionic_common import FermionicCommon
from ..fermionic_local_operators import FermionicOperator
from ..symmetries import calc_phase_permutation, get_symmetry
from ..utils import DEBUG, get_rng
from .sparse_array_common import (
    SparseArrayCommon,
    permuted,
    truncate_svd_result_blocksparse,
)
from .sparse_data_common import BlockCommon
from .sparse_vector import BlockVector


def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)


def oddpos_parse(oddpos, parity):
    """Parse the label ``oddpos``, given the parity of the array."""
    if parity and oddpos is None:
        raise ValueError(
            "`oddpos` required for calculating global phase of "
            "contractions with odd parity fermionic arrays."
        )

    if isinstance(oddpos, (list, tuple)):
        if len(oddpos) == 0:
            return ()
        elif isinstance(oddpos[0], FermionicOperator):
            # an explicit sequence of subsumed odd ranks
            return tuple(oddpos)

    if not parity:
        # we can just drop even parity labels
        return ()

    if not isinstance(oddpos, FermionicOperator):
        oddpos = FermionicOperator(oddpos)

    return (oddpos,)


class FermionicArray(
    FermionicCommon,
    SparseArrayCommon,
    BlockCommon,
    AbelianCommon,
    SymmrayCommon,
):
    """A fermionic block symmetry array.

    Parameters
    ----------
    indices : tuple of Index
        The indices of the array.
    charge : hashable, optionals
        The total charge of the array, if not given it will be inferred from
        either the first sector or set to the identity charge, if no sectors
        are given.
    blocks : dict, optional
        The blocks of the array, by default empty.
    phases : dict, optional
        The lazy phases of each block, by default empty.
    label : hashable, optional
        An optional label for the array, potentially needed for ordering dummy
        odd fermionic modes.
    symmetry : str or Symmetry, optional
        The symmetry of the array, if not using a specific symmetry class.
    """

    __slots__ = (
        "_blocks",
        "_charge",
        "_indices",
        "_label",
        "_oddpos",
        "_phases",
        "_symmetry",
    )
    fermionic = True
    static_symmetry = None

    def __init__(
        self,
        indices,
        charge=None,
        blocks=(),
        phases=(),
        label=None,
        symmetry=None,
        oddpos=None,
    ):
        self._init_abelian(
            indices=indices,
            charge=charge,
            blocks=blocks,
            symmetry=symmetry,
            label=label,
        )
        self._phases = dict(phases)

        if oddpos is None and self.label is not None:
            # default to the array label
            oddpos = self.label

        self._oddpos = oddpos_parse(oddpos, self.parity)
        if DEBUG:
            self.check()

    @property
    def parity(self):
        """The parity of the total charge."""
        return self.symmetry.parity(self.charge)

    @property
    def phases(self):
        """The lazy phases of each sector. Trivial phases are not necessarily
        stored.
        """
        try:
            return self._phases
        except AttributeError:
            self._phases = {}
            return self._phases

    @property
    def oddpos(self):
        """The odd rank of the array, i.e. the ordering of subsumed odd
        positions.
        """
        try:
            return self._oddpos
        except AttributeError:
            self._oddpos = ()
            return self._oddpos

    def copy(self):
        """Create a copy of this fermionic array.

        Returns
        -------
        FermionicArray
            The copied array.
        """
        new = self._copy_abelian()
        new._phases = self.phases.copy()
        new._oddpos = self.oddpos
        return new

    def new_with(self, indices, charge, blocks):
        """Create a new block sparse fermionic array of the same class as this
        one. Unlike `copy`, this does not copy over any existing data and drops
        for example `label`, `phases`, and `oddpos`.
        """
        new = self._new_with_abelian(
            indices=indices,
            charge=charge,
            blocks=blocks,
        )
        new._phases = {}
        new._oddpos = ()
        return new

    def copy_with(self, indices=None, blocks=None, charge=None, phases=None):
        """Create a copy of this fermionic array with some attributes replaced.

        Parameters
        ----------
        indices : tuple of Index, optional
            The new indices, if None, the original indices are used.
        blocks : dict, optional
            The new blocks, if None, the original blocks are used.
        charge : int, optional
            The new total charge, if None, the original charge is used.
        phases : dict, optional
            The new phases, if None, the original phases are used.
        """
        new = self._copy_with_abelian(
            indices=indices, blocks=blocks, charge=charge
        )
        new._phases = self.phases.copy() if phases is None else phases
        new._oddpos = self.oddpos

        if DEBUG:
            new.check()

        return new

    def modify(
        self,
        indices=None,
        blocks=None,
        charge=None,
        phases=None,
        oddpos=None,
    ):
        """Modify this fermionic array inplace. This is for internal use, and
        does not perform any checks on the updated attributes.

        Parameters
        ----------
        indices : tuple of Index, optional
            The new indices, if None, the original indices are used.
        blocks : dict, optional
            The new blocks, if None, the original blocks are used.
        charge : int, optional
            The new total charge, if None, the original charge is used.
        phases : dict, optional
            The new phases, if None, the original phases are used.
        oddpos : object or FermionicOperator, optional
            The new oddpos, if None, the original oddpos is used.
        """
        if phases is not None:
            self._phases = phases
        if oddpos is not None:
            self._oddpos = oddpos
        return self._modify_abelian(
            indices=indices, blocks=blocks, charge=charge
        )

    def randomize_phases(self, seed=None, inplace=False) -> "FermionicArray":
        """Randomize the phases of each sector to either +1 or -1. This is
        useful for testing.

        Parameters
        ----------
        seed : int or numpy.random.Generator, optional
            The random seed or generator, by default None.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
            The phase randomized array.
        """
        rng = get_rng(seed)
        new_phases = {}
        for sector in self.sectors:
            if rng.uniform() > 0.5:
                new_phases[sector] = -1
        return self._modify_or_copy(inplace=inplace, phases=new_phases)

    def phase_sync(self, inplace=False) -> "FermionicArray":
        """Multiply all lazy phases into the block arrays.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
            The resolved array, which now has no lazy phases.
        """
        new = self if inplace else self.copy()
        phases = new.phases
        while phases:
            sector, phase = phases.popitem()
            if phase == -1:
                try:
                    new.set_block(sector, -new.get_block(sector))
                except KeyError:
                    # if the block is not present, it is zero
                    # this can happen e.g. if two arrays have been aligned
                    # for contraction
                    # TODO: use a drop_sectors method instead?
                    pass

        return new

    def phase_flip(self, *axs, inplace=False) -> "FermionicArray":
        """Flip the phase of all sectors with odd parity at the given axis.

        Parameters
        ----------
        ax : int
            The axis along which to flip the phase.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
            The phase-flipped array.
        """
        new = self if inplace else self.copy()

        if not axs:
            # nothing to do
            return new

        new_phases = new.phases.copy()

        for sector in new.sectors:
            parity = sum(new.symmetry.parity(sector[ax]) for ax in axs) % 2

            if parity:
                new_phase = -new_phases.get(sector, 1)
                if new_phase == 1:
                    new_phases.pop(sector, None)
                else:
                    # only keep non-trivial phases
                    new_phases[sector] = new_phase

        return new.modify(phases=new_phases)

    def phase_transpose(self, axes=None, inplace=False):
        """Phase this fermionic array as if it were transposed virtually, i.e.
        the actual arrays are not transposed. Useful when one wants the actual
        data layout to differ from the required fermionic mode layout.

        Parameters
        ----------
        axes : tuple of int, optional
            The permutation of axes, by default None.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
        """
        new = self if inplace else self.copy()

        for sector in new.sectors:
            parities = tuple(new.symmetry.parity(q) for q in sector)

            phase_new = (
                # start with old phase
                new._phases.get(sector, 1)
                *
                # get the phase from permutation
                calc_phase_permutation(parities, axes)
            )

            if phase_new == 1:
                new._phases.pop(sector, None)
            else:
                new._phases[sector] = phase_new

        return new

    def phase_sector(self, sector, inplace=False):
        """Flip the phase of a specific sector.

        Parameters
        ----------
        sector : tuple of hashable
            The sector to flip the phase of.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
        """
        new = self if inplace else self.copy()
        new_phase = -new._phases.pop(sector, 1)
        if new_phase == -1:
            new._phases[sector] = -1
        return new

    def phase_global(self, inplace=False):
        """Flip the global phase of the array.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
        """
        new = self if inplace else self.copy()
        for sector in new.sectors:
            phase = -new.phases.pop(sector, 1)
            if phase == -1:
                new._phases[sector] = phase
        return new

    def _map_blocks(self, fn_block=None, fn_sector=None, fn_filter=None):
        """Map the blocks and their keys (sectors) of the array inplace."""
        self._map_blocks_blockcommon(fn_block, fn_sector, fn_filter)
        if fn_sector is not None:
            # need to update phase keys as well
            self._phases = {
                fn_sector(s): p
                for s, p in self._phases.items()
                if fn_filter is None or fn_filter(s)
            }

    def _resolve_oddpos_conj(self, phase_permutation=True):
        """Assuming we have effectively taken the conjugate of a fermionic
        array with dummy oddpos modes, get their new order and compute any
        phase changes coming from moving back to the beginning of the index
        order.
        """
        if not self.oddpos:
            return

        # 1. we get a reversal and conjugation of the oddpos modes
        #       dummy modes          real indices
        # | o0 o1 ... on-2 on-1 | P0 P1 ... Pn-2 Pn-1 |
        #                     <-->
        # | Pn-1 Pn-2 ... P1 P0 | on-1 on-2 ... o1 o0 |
        new_oddpos = tuple(r.dag for r in reversed(self.oddpos))
        self.modify(oddpos=new_oddpos)

        if phase_permutation and self.parity and len(new_oddpos) % 2:
            # 2. moving oddpos charges back to left
            # after flipping might generate global sign
            # | Pn-1 Pn-2 ... P1 P0 | on-1 on-2 ... o1 o0 |
            #                     <--
            # | on-1 on-2 ... o1 o0 | Pn-1 Pn-2 ... P1 P0 |
            self.phase_global(inplace=True)

    def _resolve_oddpos_combine(self, left, right):
        """Calculate the new combined dummy odd modes and any associated global
        phases combing from contracting two fermionic arrays `a` and `b`. This
        modifies this array in place.
        """
        l_oddpos = left.oddpos
        r_oddpos = right.oddpos

        if not l_oddpos and not r_oddpos:
            self._oddpos = ()
            return

        oddpos = [*l_oddpos, *r_oddpos]

        # e.g. (1, 2, 4, 5) + (3, 6, 7) -> [1, 2, 4, 5, 3, 6, 7]
        if left.parity and len(r_oddpos) % 2 == 1:
            # moving right oddpos charges over left sectors will generate sign
            phase = -1
        else:
            phase = 1

        # do a phased sort and annihilation of conjugate pairs
        i = 0
        while i < len(oddpos) - 1:
            a = oddpos[i]
            b = oddpos[i + 1]
            if a.label == b.label:
                # 'trace' out the pair
                if a.dual != b.dual:
                    if b.dual:
                        # |x><x|, ket-bra contraction
                        phase = -phase
                    oddpos.pop(i)
                    oddpos.pop(i)
                    # check previous
                    i = max(0, i - 1)
                else:
                    # detect non conjugate duplicates here as well
                    raise ValueError(
                        "`oddpos` must be unique conjugate pairs."
                    )
            elif b < a:
                # sort with phased swap
                oddpos[i] = b
                oddpos[i + 1] = a
                # check previous
                i = max(0, i - 1)
                phase = -phase
            else:
                # already sorted and not conjugate pair, move to next
                i += 1

        if phase == -1:
            self.phase_global(inplace=True)

        self._oddpos = tuple(oddpos)

    def _resolve_oddpos_squeeze(self, axes_squeeze):
        """Assuming we are about to squeeze away `axes_squeeze`, compute the
        phases associated with moving them to the beginning of the array, and
        then turn them into dummy oddpos modes.
        """
        axes_leave = []
        squeezed_oddpos = []
        for ax, ix in enumerate(self.indices):
            if ax in axes_squeeze:
                (c,) = ix._chargemap
                if self.symmetry.parity(c):
                    # if we are squeezing an odd parity charge, then we need
                    # to 'move it' into an extra dummy mode
                    label = self.label
                    if label is None:
                        raise ValueError(
                            "Cannot squeeze fermionic index with odd parity "
                            "if array has no ordering `.label` attribute."
                        )
                    op = FermionicOperator(("squeeze", label, ax), ix.dual)
                    squeezed_oddpos.append(op)
            else:
                axes_leave.append(ax)

        if squeezed_oddpos:
            # only need to update we have removed odd modes
            self.phase_transpose((*axes_squeeze, *axes_leave), inplace=True)
            self.modify(oddpos=(*self.oddpos, *squeezed_oddpos))

    def transpose(self, axes=None, phase=True, inplace=False):
        """Transpose the fermionic array, by default accounting for the phases
        accumulated from swapping odd charges.

        Parameters
        ----------
        axes : tuple of int, optional
            The permutation of axes, by default None.
        phase : bool, optional
            Whether to flip the phase of sectors whose odd charges undergo a
            odd permutation. By default True.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
            The transposed array.
        """
        new = self if inplace else self.copy()

        old_phases = new.phases

        if axes is None:
            axes = tuple(range(new.ndim - 1, -1, -1))
        else:
            axes = tuple(axes)

        if phase:
            # compute new sector phases
            new_phases = {}
            for sector in new.sectors:
                parities = tuple(new.symmetry.parity(q) for q in sector)
                perm_phase = calc_phase_permutation(parities, axes)
                new_phase = old_phases.get(sector, 1) * perm_phase
                if new_phase == -1:
                    # only populate non-trivial phases
                    new_phases[permuted(sector, axes)] = -1
        else:
            # just permute the phase keys
            new_phases = {
                permuted(sector, axes): phase
                for sector, phase in old_phases.items()
            }

        # update phase dict
        new.modify(phases=new_phases)

        # then transpose block arrays
        new._transpose_abelian(axes=axes, inplace=True)

        return new

    def conj(self, phase_permutation=True, phase_dual=False, inplace=False):
        """Conjugate this fermionic array. By default this include phases from
        both the virtual flipping of all axes, but *not* the conjugation of
        dual indices, such that::

            (
                tensordot_fermionic(x.conj(), x, ndim) ==
                tensordot_fermionic(x, x.conj(), ndim)
            )

        If all indices have matching dualness (i.e. all bra or all ket), *or*
        you set `phase_dual=True` then the above contractions will also be
        equal to ``x.norm() ** 2``.

        Parameters
        ----------
        phase_permutation : bool, optional
            Whether to flip the phase of sectors whose odd charges undergo a
            odd permutation due to *virtually* flipping the order of axes, by
            default True.
        phase_dual : bool, optional
            Whether to flip the phase of dual indices, by default False. If a
            FermionicArray has a mix of dual and non-dual indices, and you are
            explicitly forming the norm, you may want to set this to True. But
            if it is part of a large tensor network you only need to flip the
            phase of true 'outer' dual indices.

        Returns
        -------
        FermionicArray
        """
        new = self if inplace else self.copy()

        _conj = ar.get_lib_fn(new.backend, "conj")

        new_indices = tuple(ix.conj() for ix in new.indices)

        if phase_dual:
            axs_conj = tuple(
                ax for ax, ix in enumerate(new_indices) if not ix.dual
            )
        else:
            axs_conj = None

        for sector, array in new.get_sector_block_pairs():
            # conjugate the actual array
            new.set_block(sector, _conj(array))

            if phase_permutation or phase_dual:
                parities = tuple(map(new.symmetry.parity, sector))

                # start with old phase
                phase_new = new._phases.get(sector, 1)

                if phase_permutation:
                    # get the phase from 'virtually' reversing all axes:
                    #     (perm=[ndim-1, ..., 0])
                    phase_new *= calc_phase_permutation(parities, None)

                if phase_dual:
                    # get the phase from conjugating 'bra' indices
                    phase_new *= (
                        -1 if (sum(parities[ax] for ax in axs_conj) % 2) else 1
                    )

                if phase_new == 1:
                    new._phases.pop(sector, None)
                else:
                    new._phases[sector] = phase_new

        new.modify(
            indices=new_indices,
            charge=new.symmetry.sign(new._charge),
        )

        new._resolve_oddpos_conj(phase_permutation=phase_permutation)

        return new

    def dagger(self, phase_dual=False, inplace=False):
        """Fermionic adjoint, implements `.H` attribute.

        Parameters
        ----------
        phase_dual : bool, optional
            Whether to flip the phase of dual indices, by default False. If a
            FermionicArray has a mix of dual and non-dual indices, and you are
            explicitly forming the norm, you may want to set this to True. But
            if it is part of a large tensor network you only need to flip the
            phase of 'outer' dual indices.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
            The conjugate transposed array.
        """
        new = self if inplace else self.copy()

        _conj = ar.get_lib_fn(new.backend, "conj")
        _transpose = ar.get_lib_fn(new.backend, "transpose")

        new_indices = tuple(ix.conj() for ix in reversed(new.indices))

        # conjugate transpose all arrays
        new_blocks = {}
        new_phases = {}
        for sector, array in new.get_sector_block_pairs():
            new_sector = sector[::-1]

            # keep -1 phases and update dict with reversed sector
            if new._phases.pop(sector, 1) == -1:
                new_phases[new_sector] = -1

            new_blocks[new_sector] = _transpose(_conj(array))

        new.modify(
            indices=new_indices,
            charge=new.symmetry.sign(new._charge),
            blocks=new_blocks,
            phases=new_phases,
        )

        if phase_dual:
            axs_conj = tuple(
                ax for ax, ix in enumerate(new_indices) if not ix.dual
            )
            new.phase_flip(*axs_conj, inplace=True)

        # handle potential dummy odd modes
        #     dagger defined by phaseless reversal of all axes
        new._resolve_oddpos_conj(phase_permutation=True)

        return new

    # --------------------------- linalg methods ---------------------------- #
    # note: most are in fermionic_common.py

    def eigh(
        self,
        phase_eigenvalues=True,
        **kwargs,
    ) -> tuple["BlockVector", "FermionicArray"]:
        """Eigenvalue decomposition of this assumed Hermitian block sparse
        fermionic array.

        Parameters
        ----------
        phase_eigenvalues : bool, optional
            If True, any local phase will be absorbed into the eigenvalues,
            such that `U @ diag(w) @ U.H == a` always holds. By default
            True. If `False` then one of `U` or `U.H` should be phased
            individually to account for local phasesin the above expression.

        Returns
        -------
        w : BlockVector
            The eigenvalues as a vector. If `phase_eigenvalues` is True, then
            eigenvalues corresponding to odd parity sectors will be negated
            depending of the inner index dualness.
        U : FermionicArray
            The array of eigenvectors.
        """
        x = self.phase_sync()
        w, U = x._eigh_abelian(**kwargs)

        if phase_eigenvalues and (not x.indices[1].dual):
            symm = x.symmetry
            # inner index is like |x><x| so introduce a phase flip,
            # we don't explicitly have Wdag so put phase in eigenvalues
            # XXX: is this the most compatible thing to do?
            # it means U @ diag(w) @ U.H == a always
            for c in w.sectors:
                if symm.parity(c):
                    w.set_block(c, -w.get_block(c))

        return w, U

    def eigh_truncated(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        positive=0,
        drop_oddpos=True,
        **kwargs,
    ) -> tuple["FermionicArray", BlockVector, "FermionicArray"]:
        """Truncated hermitian eigen-decomposition of this assumed hermitian
        block sparse fermionic array.

        Parameters
        ----------
        cutoff : float, optional
            Absolute eigenvalue cutoff threshold.
        cutoff_mode : int or str, optional
            How to perform the truncation:

            - 1 or 'abs': trim values below ``cutoff``
            - 2 or 'rel': trim values below ``s[0] * cutoff``
            - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

        max_bond : int
            An explicit maximum bond dimension, use -1 for none.
        absorb : {-1, 0, 1, None}
            How to absorb the eigenvalues.

            - -1 or 'left': absorb into the left factor (U).
            - 0 or 'both': absorb the square root into both factors.
            - 1 or 'right': absorb into the right factor (VH).
            - None: do not absorb, return eigenvalues as a BlockVector.

        renorm : {0, 1}
            Whether to renormalize the eigenvalues (depends on `cutoff_mode`).
        positive: bool, optional
            If True, assume all eigenvalues are positive for a faster sort.
            By default False.
        drop_oddpos : bool, optional
            Whether to drop any dummy oddpos modes after the decomposition.
            By default True.

        Returns
        -------
        u : FermionicArray
            The fermionic array of left eigenvectors.
        w : BlockVector or None
            The vector of eigenvalues, or None if absorbed.
        uh : FermionicArray
            The fermionic array of right eigenvectors.
        """
        if kwargs:
            import warnings

            warnings.warn(
                f"Got unexpected kwargs {kwargs} in eigh_truncated "
                f"for {self.__class__}. Ignoring them.",
                UserWarning,
            )

        # since we are handling UH, we can add phase there
        w, U = self.eigh(phase_eigenvalues=False)

        # inplace sort by descending magnitude
        for sector, charge in zip(U.sectors, w.sectors):
            evals = w.get_block(charge)
            evecs = U.get_block(sector)

            if not positive:
                idx = ar.do(
                    "argsort",
                    -ar.do("abs", evals, like=self.backend),
                    like=self.backend,
                )
                w.set_block(charge, evals[idx])
                U.set_block(sector, evecs[:, idx])
            else:
                # assume positive, just need to flip
                w.set_block(charge, evals[::-1])
                U.set_block(sector, evecs[:, ::-1])

        if DEBUG:
            U.check()
            w.check()
            U.check_with(w, 1)

        if drop_oddpos:
            # don't propagate dummy oddpos modes
            U._oddpos = ()
            U._label = ()

        VH = U._dagger_abelian()
        if VH.indices[0].dual:
            # inner index is like |x><x| so introduce a phase flip
            VH.phase_flip(0, inplace=True)

        return truncate_svd_result_blocksparse(
            U,
            w,
            VH,
            cutoff,
            cutoff_mode,
            max_bond,
            absorb,
            renorm,
            backend=self.backend,
            use_abs=not positive,
        )


# --------------- specific fermionic symmetric array classes ---------------- #


class Z2FermionicArray(FermionicArray):
    """A fermionic block array with Z2 symmetry."""

    static_symmetry = get_symmetry("Z2")

    def to_pyblock3(self, flat=False):
        from pyblock3.algebra.fermion import SparseFermionTensor, SubTensor
        from pyblock3.algebra.fermion_symmetry import Z2

        blocks = [
            SubTensor(
                -array if self.phases.get(sector, 1) == -1 else array,
                q_labels=tuple(map(Z2, sector)),
            )
            for sector, array in self.get_sector_block_pairs()
        ]

        data = SparseFermionTensor(
            blocks,
            pattern=["-" if dual else "+" for dual in self.duals],
        )

        if flat:
            data = data.to_flat()

        data.shape = self.shape

        return data


class U1FermionicArray(FermionicArray):
    """A fermionic block array with U1 symmetry."""

    static_symmetry = get_symmetry("U1")

    def to_pyblock3(self, flat=False):
        from pyblock3.algebra.fermion import SparseFermionTensor, SubTensor
        from pyblock3.algebra.fermion_symmetry import U1

        blocks = [
            SubTensor(
                -array if self.phases.get(sector, 1) == -1 else array,
                q_labels=tuple(map(U1, sector)),
            )
            for sector, array in self.get_sector_block_pairs()
        ]

        data = SparseFermionTensor(
            blocks,
            pattern=["-" if dual else "+" for dual in self.duals],
        )

        if flat:
            data = data.to_flat()

        data.shape = self.shape

        return data


class Z2Z2FermionicArray(FermionicArray):
    """A fermionic block array with Z2 x Z2 symmetry."""

    static_symmetry = get_symmetry("Z2Z2")


class U1U1FermionicArray(FermionicArray):
    """A fermionic block array with U1 x U1 symmetry."""

    static_symmetry = get_symmetry("U1U1")



import operator

import pytest
from numpy.testing import assert_allclose

import symmray as sr


@pytest.mark.parametrize("op", ["mul", "truediv"])
def test_block_scalar_operations(op):
    fn = getattr(operator, op)
    x = sr.utils.get_rand("Z2", (7, 5, 6), duals=(1, 0, 1), charge=0)
    y = 3.14
    z = fn(x, y)
    assert_allclose(
        z.to_dense(),
        fn(x.to_dense(), y),
    )


@pytest.mark.parametrize("op", ["mul", "sub", "add", "truediv", "pow"])
@pytest.mark.parametrize("rhs", [False, True])
def test_blockvector_scalar_operations(op, rhs):
    fn = getattr(operator, op)
    x = sr.utils.get_rand_blockvector(10, block_size=3)
    xd = x.to_dense()
    y = yd = 3.14
    if rhs:
        x, y = y, x
        xd, yd = yd, xd
    z = fn(x, y)
    assert_allclose(
        z.to_dense(),
        fn(xd, yd),
    )


@pytest.mark.parametrize("op", ["mul", "sub", "add", "truediv", "pow"])
def test_blockvector_binary_operations(op):
    fn = getattr(operator, op)
    x = sr.utils.get_rand_blockvector(10, block_size=3)
    y = sr.utils.get_rand_blockvector(10, block_size=3)
    z = fn(x, y)
    assert_allclose(
        z.to_dense(),
        fn(x.to_dense(), y.to_dense()),
    )



"""Index objects for flat arrays."""

import autoray as ar

from ..index_common import Index, SubInfo
from ..utils import DEBUG


class FlatIndex(Index):
    """Simple class to store dualness and any fuse information of an index.

    Parameters
    ----------
    dual : bool, optional
        Whether the index is dual (i.e., contributes a negative sign to the
        corresponding charge). Default is False.
    subinfo : FlatSubIndexInfo, optional
        Information about the subindex, if this index is a fused index.
        Default is None, which means the index is not fused.
    """

    __slots__ = ("_num_charges", "_charge_size", "_dual", "_subinfo")

    def __init__(
        self,
        num_charges,
        charge_size,
        dual=False,
        subinfo=None,
        linearmap=None,
    ):
        self._num_charges = int(num_charges)
        self._charge_size = int(charge_size)
        self._dual = dual
        self._subinfo = subinfo
        self._linearmap = linearmap

        if DEBUG:
            self.check()

    def copy_with(
        self,
        num_charges=None,
        charge_size=None,
        dual=None,
        linearmap=None,
        **kwargs,
    ):
        """A copy of this index with some attributes replaced. Note that checks
        are not performed on the new propoerties, this is intended for internal
        use.
        """
        new = self.__new__(self.__class__)
        keep_linearmap = True

        if num_charges is None:
            new._num_charges = self._num_charges
        else:
            new._num_charges = num_charges
            keep_linearmap = False

        if charge_size is None:
            new._charge_size = self._charge_size
        else:
            new._charge_size = charge_size
            keep_linearmap = False

        if linearmap is not None:
            new._linearmap = linearmap
        elif keep_linearmap:
            new._linearmap = self._linearmap
        else:
            new._linearmap = None

        new._dual = dual if dual is not None else self._dual

        # handle subinfo to distinguish between passing None and not passing it
        if "subinfo" in kwargs:
            new._subinfo = kwargs.pop("subinfo")
        else:
            new._subinfo = self._subinfo
        if kwargs:
            raise TypeError(f"Unexpected keyword arguments: {kwargs}")

        return new

    @property
    def num_charges(self) -> int:
        """The number of charges associated with this index."""
        return self._num_charges

    @property
    def charge_size(self) -> int:
        """The size of the charges associated with this index."""
        return self._charge_size

    # @property
    # def linearmap(self):
    #     """The linear map from charge and offset to linear index."""
    #     # XXX: this should be an array compat operation
    #     if self._linearmap is None:
    #         linearmap = []
    #         for c in range(self._num_charges):
    #             for ci in range(self._charge_size):
    #                 linearmap.append((c, ci))
    #         self._linearmap = tuple(linearmap)
    #     return self._linearmap

    @property
    def subshape(self):
        if self._subinfo is None:
            return None
        return self._subinfo.subshape

    @property
    def size_total(self) -> int:
        return self._num_charges * self._charge_size

    def conj(self) -> "FlatIndex":
        """Return the conjugate of the index, i.e., flip the dualness and
        subinfo.
        """
        return FlatIndex(
            num_charges=self._num_charges,
            charge_size=self._charge_size,
            dual=not self._dual,
            subinfo=None if self._subinfo is None else self._subinfo.conj(),
        )

    def select_charge(self, charge, subselect=None) -> "FlatIndex":
        """Drop all but the specified charge from this index.

        Parameters
        ----------
        charge : int
            The charge to keep.
        subselect : slice or array_like, optional
            If provided, a range of indices within the selected charge block
            to keep. If not provided, the entire block is kept.

        Returns
        -------
        FlatIndex
        """
        updates = {"num_charges": 1}

        if subselect is not None:
            if isinstance(subselect, slice):
                start, stop, step = subselect.indices(self.charge_size)
                updates["charge_size"] = len(range(start, stop, step))
            elif hasattr(subselect, "size"):  # numpy array or similar
                updates["charge_size"] = subselect.size
            else:
                updates["charge_size"] = len(subselect)

            # drop subinfo if we slice blocks
            updates["subinfo"] = None

        elif self._subinfo is not None:
            updates["subinfo"] = self._subinfo.select_charge(charge)

        return self.copy_with(**updates)

    def linear_to_charge_and_offset(self, i):
        """Given a linear index ``i`` into this index (as if it were a dense
        array), return the corresponding charge and offset within that charge
        block.

        Parameters
        ----------
        i : int
            The linear index into this index.

        Returns
        -------
        charge : hashable
            The charge corresponding to the linear index.
        offset : int
            The offset within the charge block corresponding to the linear
            index.
        """
        if self._linearmap is None:
            # default mapping is sorted charges
            return (i // self._charge_size, i % self._charge_size)
        return self._linearmap[i]

    def check(self):
        """Check that the index is valid."""
        assert self._num_charges > 0, "Number of charges must be positive."
        assert self._charge_size > 0, "Charge size must be positive."
        if self._subinfo is not None:
            assert isinstance(self._subinfo, FlatSubIndexInfo), (
                "Subindex info must be an instance of FlatSubIndexInfo."
            )
            self._subinfo.check()

    def __repr__(self):
        s = [f"{self.__class__.__name__}("]
        s.append(
            f"num_charges={self._num_charges}, "
            f"charge_size={self._charge_size}, "
            f"dual={self._dual}"
        )
        if self._subinfo is not None:
            s.append(f", subinfo={self._subinfo!r}")
        s.append(")")
        return "".join(s)


class FlatSubIndexInfo(SubInfo):
    """Information required to unfuse a fused index.

    Parameters
    ----------
    indices : tuple[FlatIndex]
        The indices that have been fused.
    subkeys : array_like
        The subkeys for the fused index, with shape
        (ncharge, nsectors, nsubcharges). I.e. the first axis selects the
        overall fused charge, the second axis selects the subsector within
        that charge, and the third axis selects the individual charge within
        that subsector.
    """

    __slots__ = (
        "_indices",
        "_subkeys",
        "_ncharge",
        "_nsectors",
        "_nsubcharges",
    )

    def __init__(self, indices, subkeys):
        self._indices = tuple(
            x if isinstance(x, FlatIndex) else FlatIndex(x) for x in indices
        )
        self._subkeys = subkeys
        subkey_shape = ar.do("shape", subkeys)
        # number of overall charges, e.g. {0, 1} -> 2
        # number of subsectors e.g. [000, 011, 101, 110] -> 4
        # number of subcharges, e.g. 3 for above
        self._ncharge, self._nsectors, self._nsubcharges = subkey_shape

        if DEBUG:
            self.check()

    @property
    def subshape(self) -> tuple[int]:
        """The subshape of the fused index."""
        return tuple(ix.size_total for ix in self._indices)

    @property
    def subkeys(self):
        """The subkeys for the fused index."""
        return self._subkeys

    @property
    def ncharge(self) -> int:
        """Number of overall charges in this subindex."""
        return self._ncharge

    @property
    def nsectors(self) -> int:
        """Number of subsectors in this subindex."""
        return self._nsectors

    @property
    def nsubcharges(self) -> int:
        """Number of subcharges in this subindex."""
        return self._nsubcharges

    def check(self):
        assert len(self._indices) == len(self._subkeys[0, 0])
        for ix in self._indices:
            ix.check()

    def conj(self):
        return FlatSubIndexInfo(
            indices=tuple(ix.conj() for ix in self._indices),
            subkeys=self._subkeys,
        )

    def select_charge(self, charge):
        new_subkeys = self.subkeys[(charge,), ...]
        return FlatSubIndexInfo(
            indices=self.indices,
            subkeys=new_subkeys,
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"ncharge={self._ncharge}, "
            f"nsectors={self._nsectors}, "
            f"nsubcharges={self._nsubcharges})"
        )



"""Methods that apply to all Abelian symmetric arrays, regardless of
backend.
"""

import functools
import numbers
import operator

import autoray as ar

from .index_common import Index
from .symmetries import Symmetry, get_symmetry
from .utils import DEBUG, lazyabstractmethod


@functools.lru_cache(maxsize=2**15)
def calc_reshape_args(shape, newshape, subshapes):
    """Given a current block sparse shape ``shape`` a target shape ``newshape``
    and current sub index sizes ``subshapes`` (i.e. previously fused dimensions)
    compute the sequence of axes to unfuse, fuse and expand to reshape the
    array.

    Parameters
    ----------
    shape : tuple[int]
        The current shape of the array.
    newshape : tuple[int]
        The target shape of the array.
    subshapes : tuple[None or tuple[int]]
        The sizes of the subindices that were previously fused.

    Returns
    -------
    axs_unfuse : tuple[int]
        The axes to unfuse.
    axs_fuse : tuple[tuple[tuple[int]]]
        The axes (after unfusing) to fuse, grouped by contiguous groups.
    axs_expand : tuple[int]
        The axes (after unfusing and fusing) to expand.
    """
    # tracks position in input shape
    i = 0
    # tracks position in output shape
    j = 0
    # tracks position in post-fuse / pre-expand shape
    k = 0

    ndim_old = len(shape)
    ndim_new = len(newshape)

    term = []  # dnyamically updated labelled dimensions
    axs_squeeze = []
    unfuse_sizes = {}
    fuse_sizes = {}
    axs_expand = []
    any_singleton = False
    any_fused = False

    while i < ndim_old and j < ndim_new:
        di = shape[i]
        dj = newshape[j]

        if (subshapes[i] is not None) and (
            subshapes[i] == newshape[j : j + len(subshapes[i])]
        ):
            # unfuse, check first
            label = f"u{len(unfuse_sizes)}"
            s = 0
            for ds in subshapes[i]:
                dj = newshape[j]
                if ds != dj:
                    raise ValueError("Shape mismatch for unfuse.")
                s += 1
                j += 1
                k += 1
            unfuse_sizes[label] = s
            term.append(label)
            i += 1
        elif di == dj:
            # output dimension already
            term.append("o")
            i += 1
            j += 1
            k += 1
        elif di == 1:
            # have to handle squeezed dimensions after unfusing
            term.append("s")
            axs_squeeze.append(i)
            any_singleton = True
            i += 1
        elif dj == 1:
            # record expansion location relative to *post* fuse shape
            axs_expand.append(k)
            j += 1
        elif di < dj:
            # need to fuse
            label = f"g{len(fuse_sizes)}"
            term.append(label)
            s = 1
            i += 1
            while di < dj:
                di *= shape[i]
                term.append(label)
                i += 1
                s += 1
            if di != dj:
                raise ValueError("Shape mismatch for fuse.")
            fuse_sizes[label] = s
            any_fused = True
            j += 1
            k += 1
        else:
            raise ValueError("Shape mismatch.")

    # check trailing dimensions, which should be size 1
    for i in range(i, ndim_old):
        any_singleton = True
        term.append("s")
    for j in range(j, ndim_new):
        axs_expand.append(k)

    # first we handle unfusings
    axs_unfuse = []
    for label, s in unfuse_sizes.items():
        ax = term.index(label)
        axs_unfuse.append(ax)
        term = term[:ax] + ["o"] * s + term[ax + 1 :]

    # handle squeezes by converting them into fuse groups
    if any_singleton:
        i = 0
        label = term[i]
        if label == "s":
            # if we have squeeze axes on left, we have to group into right
            while label == "s":
                i += 1
                label = term[i]

            if label[0] == "g":
                # adjacent to existing group
                g = label
            elif label == "o":
                # or new group
                g = f"g{len(fuse_sizes)}"
                term[i] = g
                fuse_sizes[g] = 1

            # mark all axs up to this point
            for j in range(0, i):
                fuse_sizes[g] += 1
                term[j] = g

        # process rest of term, now preferring grouping into left
        i += 1
        while i < len(term):
            label = term[i]
            if label == "s":
                left = term[i - 1]
                if left[0] == "g":
                    g = left
                elif left == "o":
                    g = f"g{len(fuse_sizes)}"
                    term[i - 1] = g
                    fuse_sizes[g] = 1

                # update any right block of squeeze axs to g
                while label == "s":
                    term[i] = g
                    fuse_sizes[g] += 1
                    i += 1
                    if i == len(term):
                        break
                    label = term[i]
            i += 1

    # now we handle fusing
    axs_fuse = []
    if any_fused or any_singleton:
        # complexity here is we want to simulteneously fuse adjacent groups for
        # efficiency, but also need to handle non-adjacent groups
        current_groups = []
        i = 0
        while i < len(term):
            label = term[i]
            if label not in fuse_sizes:
                if current_groups:
                    # start of groups
                    i0 = i - sum(map(len, current_groups))
                    ng = len(current_groups)
                    term = term[:i0] + ["o"] * ng + term[i:]
                    axs_fuse.append(tuple(current_groups))
                    current_groups = []
                    # rewind to end of new group(s)
                    i = i0 + ng
                else:
                    i += 1
                continue

            s = fuse_sizes[label]
            current_groups.append(tuple(range(i, i + s)))
            i += s
        if current_groups:
            axs_fuse.append(tuple(current_groups))

    # handle expansion
    axs_expand.reverse()

    return tuple(axs_unfuse), tuple(axs_fuse), tuple(axs_expand)


class AbelianCommon:
    """Common base class for arrays with Abelian symmetry."""

    @property
    def symmetry(self) -> Symmetry:
        """The symmetry object of the array."""
        return self._symmetry

    @property
    def indices(self) -> tuple[Index, ...]:
        """The indices of the array."""
        return self._indices

    def is_fused(self, ax: int) -> bool:
        """Does axis `ax` carry subindex information, i.e., is it a fused
        index?
        """
        return self._indices[ax].subinfo is not None

    @property
    def duals(self) -> tuple[bool, ...]:
        """The dual-ness of each index."""
        return tuple(ix.dual for ix in self._indices)

    @classmethod
    def get_class_symmetry(cls, symmetry=None) -> Symmetry:
        if symmetry is None:
            if cls.static_symmetry is None:
                # symmetry must be given if not static
                raise ValueError("Symmetry must be given.")
            symmetry = cls.static_symmetry
        elif cls.static_symmetry and symmetry != cls.static_symmetry:
            raise ValueError("Cannot override static symmetry of class.")

        return get_symmetry(symmetry)

    @property
    def signature(self) -> str:
        return "".join("-" if f else "+" for f in self.duals)

    def __add__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            # can add arrays with matching sparsity and structure
            return self._binary_blockwise_op(
                other,
                fn=operator.add,
                missing="outer",
                inplace=inplace,
            )

        # addition with anything else incld scalars breaks sparsity
        raise NotImplementedError(
            f"Addition with {type(self)} and {type(other)} not implemented."
        )

    def __iadd__(self, other):
        return self.__add__(other, inplace=True)

    def __sub__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            # can subtract arrays with matching sparsity and structure
            return self._binary_blockwise_op(
                other,
                fn=operator.sub,
                inplace=inplace,
            )

        # subtraction with anything else incld scalars breaks sparsity
        raise NotImplementedError(
            f"Subtraction with {type(self)} and {type(other)} not implemented."
        )

    def __isub__(self, other):
        return self.__sub__(other, inplace=True)

    def __truediv__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            if self.shape == other.shape and all(d == 1 for d in self.shape):
                # print("when does this happen??")
                return self._binary_blockwise_op(
                    other, fn=operator.truediv, inplace=inplace
                )
            # deviding by implicit zeros not defined
            return NotImplemented

        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Division {self} / {other} not supported.")

        # assume scalar
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x / other)
        return new

    def __itruediv__(self, other):
        return self.__truediv__(other, inplace=True)

    @lazyabstractmethod
    def transpose(self, axes=None, inplace=False) -> "AbelianCommon":
        pass

    @property
    def T(self) -> "AbelianCommon":
        """The transpose of the block array."""
        return self.transpose()

    @lazyabstractmethod
    def conj(self, inplace=False) -> "AbelianCommon":
        pass

    def _dagger_abelian(self, inplace=False) -> "AbelianCommon":
        """Conjugate transpose or adjoint, implements the .H property."""
        new = self._transpose_abelian(inplace=inplace)
        new._conj_abelian(inplace=True)
        return new

    @property
    def H(self) -> "AbelianCommon":
        return self.dagger()

    @lazyabstractmethod
    def _fuse_core(self, *axes_groups, inplace=False, **kwargs):
        pass

    @lazyabstractmethod
    def expand_dims(self, ax, inplace=False) -> "AbelianCommon":
        pass

    def fuse(
        self,
        *axes_groups,
        expand_empty=True,
        inplace=False,
        **kwargs,
    ) -> "AbelianCommon":
        """Fuse the given group or groups of axes. The new fused axes will be
        inserted at the minimum index of any fused axis (even if it is not in
        the first group). For example, ``x.fuse([5, 3], [7, 2, 6])`` will
        produce an array with axes like::

            groups inserted at axis 2, removed beyond that.
                   ......<--
            (0, 1, g0, g1, 4, 8, ...)
                   |   |
                   |   g1=(7, 2, 6)
                   g0=(5, 3)

        The fused axes will carry subindex information, which can be used to
        automatically unfuse them back into their original components.
        Depending on `expand_empty`, any empty groups can be expanded to new
        singlet dimensions, or simply ignored.

        Parameters
        ----------
        axes_groups : Sequence[Sequence[int]]
            The axes to fuse. Each group of axes will be fused into a single
            axis.
        expand_empty : bool, optional
            Whether to expand empty groups into new axes.
        mode : "auto", "insert", "concat", optional
            The method to use for fusing. `"insert"` creates the new fused
            blocks and insert the subblocks inplace. `"concat"` recursively
            concatenates the subblocks, which can be slightly slower but is
            more compatible with e.g. autodiff. `"auto"` will use `"insert"` if
            the backend is numpy, otherwise `"concat"`.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.
        kwargs : dict, optional
            Additional keyword arguments to pass to the core fusing method.

        Returns
        -------
        AbelianCommon
        """
        # handle empty groups and ensure hashable
        _axes_groups = []
        _axes_expand = []
        for ax, group in enumerate(axes_groups):
            if group:
                _axes_groups.append(tuple(group))
            else:
                _axes_expand.append(ax)

        if _axes_groups:
            xf = self._fuse_core(*_axes_groups, inplace=inplace, **kwargs)
        else:
            xf = self if inplace else self.copy()

        if expand_empty and _axes_expand:
            g0 = min(g for groups in _axes_groups for g in groups)
            for ax in _axes_expand:
                xf.expand_dims(g0 + ax, inplace=True)

        return xf

    @lazyabstractmethod
    def unfuse(self, ax, inplace=False) -> "AbelianCommon":
        pass

    def unfuse_all(self, inplace=False) -> "AbelianCommon":
        """Unfuse all indices that carry subindex information, likely from a
        fusing operation.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianCommon
        """
        new = self if inplace else self.copy()
        for ax in reversed(range(self.ndim)):
            if new.is_fused(ax):
                new.unfuse(ax, inplace=True)
        return new

    def reshape(self, newshape, inplace=False) -> "AbelianCommon":
        """Reshape this abelian array to ``newshape``, assuming it can be done
        by any mix of fusing, unfusing, and expanding new axes.

        Restrictions and complications vs normal array reshaping arise from the
        fact that

            A) only previously fused axes can be unfused, and their total size
               may not be the product of the individual sizes due to sparsity.
            B) the indices carry information beyond size and how they are
               grouped potentially matters, relevant for singleton dimensions.

        Accordingly the approach here is as follows:

            1. Unfuse any axes that match the new shape.

            2. If there are singleton dimensions that don't appear in the new
               shape, (i.e. are being 'squeezed') these are grouped with the
               axis to the their left to then be fused. If they are already
               left-most, they are grouped with the right.

            3. Fuse any groups of axes required to match the new shape.
               Adjacent groups are fused simultaneously for efficiency.

            4. Expand new axes required to match singlet dimensions in the new
               shape. By default these will have zero charge and dual-ness
               iherited from whichever axis is to their left, or right if they
               are the left-most axis already.

        To avoid the effective grouping of 'squeezed' axes you can explicitly
        squeeze them before reshaping. Similarly use ``expand_dims`` to add
        new axes with specific charges and dual-ness.

        Parameters
        ----------
        newshape : tuple[int]
            The new shape to reshape to.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianCommon

        See Also
        --------
        fuse, unfuse, squeeze, expand_dims
        """
        x = self if inplace else self.copy()

        if not isinstance(newshape, tuple):
            newshape = tuple(newshape)
        newshape = ar.lazy.core.find_full_reshape(newshape, self.size)

        subshapes = tuple(ix.subshape for ix in x.indices)

        # cached parsing of reshape arguments
        axs_unfuse, axs_fuse_groupings, axs_expand = calc_reshape_args(
            x.shape, newshape, subshapes
        )

        for ax in axs_unfuse:
            x.unfuse(ax, inplace=True)
        for grouping in axs_fuse_groupings:
            x.fuse(*grouping, inplace=True)
        for ax in axs_expand:
            x.expand_dims(ax, inplace=True)

        return x

    def __str__(self):
        lines = [
            (
                f"{self.__class__.__name__}(ndim={self.ndim}, "
                f"charge={self.charge}, indices=["
            )
        ]
        for i in range(self.ndim):
            lines.extend(
                f"    {line}" for line in str(self.indices[i]).split("\n")
            )
        lines.append(
            (
                f"], num_blocks={self.num_blocks}, backend={self.backend}, "
                f"dtype={self.dtype})"
            )
        )
        return "\n".join(lines)

    @lazyabstractmethod
    def get_any_array(self):
        pass

    def __getitem__(self, item):
        axis = None
        idx = None

        if not isinstance(item, tuple):
            raise TypeError(
                f"Expected a tuple for indexing, got {type(item)}: {item}"
            )

        for i, s in enumerate(item):
            if isinstance(s, slice):
                if not s.start is s.stop is s.step is None:
                    raise NotImplementedError("Can only slice whole axes.")
            else:
                if axis is not None:
                    raise ValueError(
                        "Can only index one axis at a time, "
                        f"got {item} with multiple indices."
                    )
                axis = i
                idx = s

        return self.isel(axis, idx)

    def take(self, indices, axis, inplace=False):
        if isinstance(indices, numbers.Integral):
            return self.isel(axis, indices, inplace=inplace)
        raise NotImplementedError("Only single index selection is supported.")

    def __repr__(self):
        if self.static_symmetry is not None:
            c = f"{self.__class__.__name__}("
        else:
            c = f"{self.__class__.__name__}{self.symmetry}("

        return "".join(
            [
                c,
                (
                    f"shape~{self.shape}:[{self.signature}]"
                    if self.indices
                    else f"{self.get_any_array()}"
                ),
                f", charge={self.charge}",
                f", num_blocks={self.num_blocks})",
            ]
        )


def without(it, remove):
    """Return a tuple of the elements in ``it`` with those at positions
    ``remove`` removed.
    """
    return tuple(el for i, el in enumerate(it) if i not in remove)


def parse_tensordot_axes(axes, ndim_a, ndim_b):
    """Parse the axes argument for single integer and also negative indices.
    Returning the 4 axes groups that can be used for fusing.
    """
    if isinstance(axes, numbers.Integral):
        axes_a = tuple(range(ndim_a - axes, ndim_a))
        axes_b = tuple(range(0, axes))
    else:
        axes_a, axes_b = axes
        axes_a = tuple(x % ndim_a for x in axes_a)
        axes_b = tuple(x % ndim_b for x in axes_b)
        if not len(axes_a) == len(axes_b):
            raise ValueError("Axes must have same length.")

    # axes left on the left and right tensors respectively
    left_axes = without(range(ndim_a), axes_a)
    right_axes = without(range(ndim_b), axes_b)

    return left_axes, axes_a, axes_b, right_axes



"""Common interfaces for all symmray array objects."""

import operator

import autoray as ar

from .utils import lazyabstractmethod

_symmray_namespace = None


class SymmrayCommon:
    """Common functionality for all symmray array like objects."""

    def __array_namespace__(self, api_version=None):
        """Return the namespace for the symmray module."""
        global _symmray_namespace
        if _symmray_namespace is None:
            import symmray

            _symmray_namespace = symmray
        return _symmray_namespace

    @lazyabstractmethod
    def copy(self) -> "SymmrayCommon":
        pass

    @lazyabstractmethod
    def copy_with(self, **kwargs) -> "SymmrayCommon":
        pass

    @lazyabstractmethod
    def modify(self, **kwargs) -> "SymmrayCommon":
        pass

    def _modify_or_copy(self, inplace=False, **kwargs) -> "SymmrayCommon":
        if inplace:
            return self.modify(**kwargs)
        else:
            return self.copy_with(**kwargs)

    @lazyabstractmethod
    def item(self):
        pass

    def __float__(self):
        return float(self.item())

    def __complex__(self):
        return complex(self.item())

    def __int__(self):
        return int(self.item())

    def __bool__(self):
        return bool(self.item())

    @lazyabstractmethod
    def _binary_blockwise_op(
        self, other, fn, missing=None, inplace=False
    ) -> "SymmrayCommon":
        pass

    @lazyabstractmethod
    def apply_to_arrays(self, fn):
        pass

    def __mul__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            return self._binary_blockwise_op(
                other, fn=operator.mul, missing="inner", inplace=inplace
            )
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x * other)
        return new

    def __imul__(self, other):
        return self.__mul__(other, inplace=True)

    def __rmul__(self, other):
        return self * other

    def __neg__(self) -> "SymmrayCommon":
        new = self.copy()
        new.apply_to_arrays(operator.neg)
        return new

    def _do_unary_op(self, fn, inplace=False) -> "SymmrayCommon":
        """Perform a unary operation on blocks of the array."""
        new = self if inplace else self.copy()
        if isinstance(fn, str):
            fn = ar.get_lib_fn(self.backend, fn)
        new.apply_to_arrays(fn)
        return new

    def abs(self):
        """Get the absolute value of all elements in the array."""
        return self._do_unary_op("abs")

    def isfinite(self):
        """Check if all elements in the array are finite."""
        return self._do_unary_op("isfinite")

    def sqrt(self):
        """Get the square root of all elements in the array."""
        return self._do_unary_op("sqrt")

    def clip(self, a_min, a_max):
        """Clip the values in the array."""
        new = self.copy()
        _clip = ar.get_lib_fn(self.backend, "clip")
        new.apply_to_arrays(lambda x: _clip(x, a_min, a_max))
        return new

    @lazyabstractmethod
    def _do_reduction(self, fn):
        pass

    def max(self):
        """Get the maximum element from any block in the array."""
        return self._do_reduction("max")

    def min(self):
        """Get the minimum element from any block in the array."""
        return self._do_reduction("min")

    def sum(self):
        """Get the sum of all elements in the array."""
        return self._do_reduction("sum")

    def all(self):
        """Check if all elements in the array are True."""
        return self._do_reduction("all")

    def any(self):
        """Check if any element in the array is True."""
        return self._do_reduction("any")

    @property
    def real(self):
        """Return the real part of the array."""
        return self._do_unary_op("real")

    @property
    def imag(self):
        """Return the imaginary part of the array."""
        return self._do_unary_op("imag")



"""Definitions of various symmetries."""

import functools
from abc import ABC, abstractmethod

from .utils import get_rng


class Symmetry(ABC):
    __slots__ = ()

    @abstractmethod
    def valid(self, *charges):
        """Check if all charges are valid for the symmetry."""
        raise NotImplementedError

    @abstractmethod
    def combine(self, *charges):
        """Combine / add charges according to the symmetry."""
        raise NotImplementedError

    @abstractmethod
    def sign(self, charge, dual=True):
        """Negate a charge according to the symmetry and flag ``dual``."""
        raise NotImplementedError

    @abstractmethod
    def parity(self, charge):
        """Return the parity, 0 or 1, of a charge according to the symmetry."""
        raise NotImplementedError

    @abstractmethod
    def random_charge(self, seed=None):
        """Return a random valid charge, for testing purposes."""
        raise NotImplementedError

    def __str__(self):
        return self.__class__.__name__

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        return self.__class__ == other.__class__

    def __hash__(self):
        return hash(self.__class__)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"


@functools.lru_cache(2**14)
def sign_scalar(charge, dual=True):
    if dual:
        return -charge
    return charge


@functools.lru_cache(2**14)
def sign_tuple(charge, dual=True):
    return tuple(sign_scalar(c, dual) for c in charge)


class ZN(Symmetry):
    __slots__ = ("N",)

    def valid(self, *charges: int) -> bool:
        return all(charge in range(self.N) for charge in charges)

    def combine(self, *charges: int) -> int:
        return sum(charges) % self.N

    def sign(self, charge: int, dual=True) -> int:
        if dual:
            return (self.N - charge) % self.N
        return charge

    def parity(self, charge: int) -> int:
        return charge % 2

    def random_charge(self, seed=None) -> int:
        rng = get_rng(seed)
        return int(rng.integers(0, self.N))

    def __reduce__(self):
        return (get_symmetry, (f"Z{self.N}",))


class Z2(ZN):
    N = 2

    def sign(self, charge: int, dual=True) -> int:
        # shortcut: Z2 is self-inverse
        return charge


@functools.cache
def get_zn_symmetry_cls(N: int) -> type:
    """Get a ZN symmetry class."""
    if N == 2:
        return Z2
    return type(f"Z{N}", (ZN,), {"N": N})


class U1(Symmetry):
    __slots__ = ()

    def valid(self, *charges: int) -> bool:
        return all(isinstance(charge, int) for charge in charges)

    def combine(self, *charges: int) -> int:
        return sum(charges)

    def sign(self, charge: int, dual=True) -> int:
        return sign_scalar(charge, dual)

    def parity(self, charge: int) -> int:
        return charge % 2

    def random_charge(self, seed=None) -> int:
        rng = get_rng(seed)
        return int(rng.integers(-1, 2))


class Z2Z2(Symmetry):
    __slots__ = ()

    def valid(self, *charges: tuple[int, int]) -> bool:
        return all(
            isinstance(charge, tuple)
            and charge[0] in {0, 1}
            and charge[1] in {0, 1}
            for charge in charges
        )

    def combine(self, *charges: tuple[int, int]) -> tuple[int, int]:
        c0, c1 = 0, 0
        for cl, cr in charges:
            c0 ^= cl
            c1 ^= cr
        return (c0, c1)

    def sign(self, charge: tuple[int, int], dual=True) -> tuple[int, int]:
        # Z2Z2 is self-inverse
        return charge

    def parity(self, charge: tuple[int, int]) -> int:
        return charge[0] ^ charge[1]

    def random_charge(self, seed=None) -> tuple[int, int]:
        rng = get_rng(seed)
        charge = rng.integers(0, 2, size=2)
        return tuple(map(int, charge))


class U1U1(Symmetry):
    __slots__ = ()

    def valid(self, *charges: tuple[int, int]) -> bool:
        return all(
            isinstance(charge, tuple)
            and isinstance(charge[0], int)
            and isinstance(charge[1], int)
            for charge in charges
        )

    def combine(self, *charges: tuple[int, int]) -> tuple[int, int]:
        c0, c1 = 0, 0
        for cl, cr in charges:
            c0 += cl
            c1 += cr
        return (c0, c1)

    def sign(self, charge: tuple[int, int], dual=True) -> tuple[int, int]:
        return sign_tuple(charge, dual)

    def parity(self, charge: tuple[int, int]) -> int:
        return (charge[0] + charge[1]) % 2

    def random_charge(self, seed=None) -> tuple[int, int]:
        rng = get_rng(seed)
        charge = rng.integers(-1, 2, size=2)
        return tuple(map(int, charge))


@functools.lru_cache(2**14)
def get_symmetry(symmetry: str | Symmetry) -> Symmetry:
    """Get a symmetry instance by name.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry to get.

    Returns
    -------
    Symmetry
        The symmetry instance.
    """
    import re

    if isinstance(symmetry, Symmetry):
        return symmetry
    elif symmetry == "Z2":
        return Z2()
    elif m := re.fullmatch(r"Z(\d+)", symmetry):
        N = int(m.group(1))
        return get_zn_symmetry_cls(N)()
    elif symmetry == "U1":
        return U1()
    elif symmetry == "Z2Z2":
        return Z2Z2()
    elif symmetry == "U1U1":
        return U1U1()
    else:
        raise ValueError(f"Unknown symmetry: {symmetry}")


@functools.lru_cache(maxsize=2**15)
def calc_phase_permutation(
    parities: tuple[int, ...],
    perm: tuple[int, ...] = None,
) -> int:
    """Given sequence of parities and a permutation, compute the phase of the
    permutation acting on the odd charges. I.e. whether the number of swaps
    of odd sectors is even or odd.

    Parameters
    ----------
    parities : tuple of int
        The parities of the sectors.
    perm : tuple of int, optional
        The permutation of axes, by default None, which flips all axes.

    Returns
    -------
    int
        The phase of the permutation, either 1 or -1.
    """
    if perm is None:
        # assume flipping all
        if sum(parities) // 2 % 2:
            return -1
        return 1

    moved = set()
    swaps = 0
    for ax in perm:
        # we are moving charge at ax to the beginning
        if parities[ax]:
            # if it is odd, count how many odd charges it crosses
            for other_ax in range(ax):
                if other_ax not in moved and parities[other_ax]:
                    swaps += 1
        moved.add(ax)

    if swaps % 2:
        return -1
    return 1



import autoray as ar
import pytest
from numpy.testing import assert_allclose

import symmray as sr


@pytest.mark.parametrize("symmetry", ("Z2", "U1"))
@pytest.mark.parametrize("d0", [3, 4])
@pytest.mark.parametrize("d1", [2, 5])
@pytest.mark.parametrize("f0", [False, True])
@pytest.mark.parametrize("f1", [False, True])
@pytest.mark.parametrize("c", [0, 1])
def test_qr_basics(symmetry, d0, d1, f0, f1, c):
    x = sr.utils.get_rand(symmetry, (d0, d1), duals=[f0, f1], charge=c)
    x.check()
    q, r = sr.linalg.qr(x)
    q.check()
    r.check()
    assert (q @ r).allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "U1U1", "Z2Z2"))
@pytest.mark.parametrize("seed", range(5))
def test_qr_with_expand_dims(symmetry, seed):
    x = sr.utils.get_rand(
        symmetry,
        [4, 5, 6],
        subsizes="maximal",
        seed=seed,
    )
    y = x.reshape(
        (
            1,
            4 * 5 * 6,
        )
    )
    q, r = sr.linalg.qr(y)
    z = (q @ r).reshape((4, 5, 6))
    assert z.allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1"))
@pytest.mark.parametrize("d0", [3, 4])
@pytest.mark.parametrize("d1", [2, 5])
@pytest.mark.parametrize("f0", [False, True])
@pytest.mark.parametrize("f1", [False, True])
@pytest.mark.parametrize("c", [0, 1])
def test_svd_basics(symmetry, d0, d1, f0, f1, c):
    x = sr.utils.get_rand(
        symmetry,
        (d0, d1),
        duals=[f0, f1],
        charge=c,
        subsizes="maximal",
    )
    x.check()
    u, s, vh = sr.linalg.svd(x)
    u.check()
    s.check()
    vh.check()
    us = ar.do("multiply_diagonal", u, s, axis=1)
    usvh = sr.tensordot(us, vh, 1)
    assert usvh.allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("d", (2, 3, 4, 5, 7))
def test_eigh(symmetry, d):
    x = sr.utils.get_rand(
        symmetry,
        (d, d),
        duals=[0, 1],
        subsizes="equal",
    )
    # need to make sure x is hermitian
    x.apply_to_arrays(lambda x: (x + x.T) / 2)
    el, ev = ar.do("linalg.eigh", x)
    el.check()
    ev.check()
    xr = ev @ ar.do("multiply_diagonal", ev.H, el, axis=0)
    xr.check()
    assert x.allclose(xr)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("d", (2, 7, 31))
def test_solve(symmetry, d):
    ind = sr.utils.rand_index(symmetry, d)

    a = sr.utils.get_rand(
        symmetry,
        shape=(ind, ind.conj()),
        dtype="complex128",
    )
    b = sr.utils.get_rand(
        symmetry,
        shape=[ind],
        dtype="complex128",
    )
    x = sr.linalg.solve(a, b)
    x.check()
    assert (a @ x).allclose(b)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("d0", (2, 7, 31))
@pytest.mark.parametrize("d1", (2, 3, 4))
def test_solve_2d(symmetry, d0, d1):
    ind = sr.utils.rand_index(symmetry, d0)
    ind1 = sr.utils.rand_index(symmetry, d1)

    a = sr.utils.get_rand(
        symmetry,
        shape=(ind, ind.conj()),
        dtype="complex128",
    )
    b = sr.utils.get_rand(
        symmetry,
        shape=(ind, ind1),
        dtype="complex128",
    )
    x = sr.linalg.solve(a, b)
    x.check()
    assert (a @ x).allclose(b)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("d", (2, 3, 4, 5, 7))
def test_expm_with_reshape(symmetry, d):
    pytest.importorskip("scipy")

    x = sr.utils.get_rand(
        symmetry,
        (d, d, d, d),
        duals=[0, 0, 1, 1],
        subsizes="equal",
    )
    x_matrix = ar.do("reshape", x, (d**2, d**2))
    # == x_matrix = x.fuse((0, 1), (2, 3))
    xe_matrix = ar.do("scipy.linalg.expm", x_matrix)
    xe_matrix.check()
    xe = ar.do("reshape", xe_matrix, (d, d, d, d))
    xe.check()
    #  == xe = xe_matrix.unfuse_all()
    xe_dense = ar.do(
        "scipy.linalg.expm", x.to_dense().reshape((d**2, d**2))
    ).reshape((d, d, d, d))
    assert_allclose(xe.to_dense(), xe_dense)



"""Basic classes and functions for block sparse backend objects."""

import functools
import operator

import autoray as ar

from ..utils import lazyabstractmethod


def _identity(x):
    return x


class BlockCommon:
    """Mixin class for arrays consisting of dicts of blocks."""

    __slots__ = ("_blocks",)

    def _init_blockcommon(self, blocks):
        self._blocks = dict(blocks)

    def _new_with_blockcommon(self, blocks):
        new = self.__new__(self.__class__)
        new._blocks = blocks
        return new

    def _copy_blockcommon(self):
        new = self.__new__(self.__class__)
        new._blocks = self._blocks.copy()
        return new

    @lazyabstractmethod
    def copy(self):
        pass

    def _copy_with_blockcommon(self, blocks=None):
        new = self.__new__(self.__class__)
        new._blocks = self._blocks.copy() if blocks is None else blocks
        return new

    @lazyabstractmethod
    def copy_with(self, blocks=None):
        pass

    def _modify_blockcommon(self, blocks=None):
        if blocks is not None:
            self._blocks = blocks
        return self

    @lazyabstractmethod
    def modify(self, blocks=None):
        pass

    @property
    def blocks(self):
        """The blocks of the array."""
        return self._blocks

    def get_block(self, sector):
        """Get the block for the given sector."""
        return self._blocks[sector]

    def set_block(self, sector, block):
        """Set the block for the given sector."""
        self._blocks[sector] = block

    def del_block(self, sector):
        """Delete the block for the given sector."""
        del self._blocks[sector]

    def get_sector_block_pairs(self):
        """Get an iterator over all `(sector, block)` pairs."""
        return self._blocks.items()

    def get_all_blocks(self):
        """Get an iterator over all blocks of the array."""
        return self._blocks.values()

    def _map_blocks_blockcommon(
        self,
        fn_block=None,
        fn_sector=None,
        fn_filter=None,
    ):
        """Map the blocks and their keys (sectors) of the array inplace."""
        if fn_block is None:
            fn_block = _identity

        if fn_sector is None:
            fn_sector = _identity

        if fn_filter is None:

            def fn_filter(sector):
                return True

        self._blocks = {
            fn_sector(sector): fn_block(block)
            for sector, block in self.get_sector_block_pairs()
            if fn_filter(sector)
        }

    @lazyabstractmethod
    def _map_blocks(self, fn_block=None, fn_sector=None, fn_filter=None):
        pass

    def get_any_sector(self):
        return next(iter(self._blocks.keys()), ())

    def get_any_array(self):
        """Get any array from the blocks, to check type and backend for
        example.
        """
        return next(iter(self.get_all_blocks()), 0.0)

    def get_scalar_element(self):
        """Assuming the block array is a scalar, get that scalar element."""
        sectors = self.sectors

        if sectors == ((),):
            # single scalar block, usual case
            return self.get_block(())
        elif sectors == ():
            # no aligned blocks, return zero
            return 0.0
        else:
            raise ValueError("Block array does not seem to be a scalar.")

    def is_zero(self, tol=1e-12):
        """Check if all blocks are zero up to a tolerance."""
        return all(
            ar.do("allclose", b, 0.0, atol=tol) for b in self.get_all_blocks()
        )

    @property
    def dtype(self):
        """Get the dtype name from an arbitrary block in the array."""
        return ar.get_dtype_name(self.get_any_array())

    @property
    def backend(self):
        """Get the backend name from an arbitrary block in the array."""
        return ar.infer_backend(self.get_any_array())

    @property
    def num_blocks(self):
        """The number of blocks in the array."""
        return len(self._blocks)

    @property
    def sectors(self):
        """Get the sectors, i.e. keys of the blocks."""
        return tuple(self._blocks.keys())

    def has_sector(self, sector):
        """Check if the array has a block for the given sector."""
        return sector in self._blocks

    def get_params(self):
        """Get the parameters of this block array as a pytree (dict).

        Returns
        -------
        dict[tuple, array_like]
        """
        return self._blocks.copy()

    def set_params(self, params):
        """Set the parameters of this block array from a pytree (dict).

        Parameters
        ----------
        params : dict[tuple, array_like]
        """
        self._blocks.update(params)

    def apply_to_arrays(self, fn):
        """Apply the ``fn`` inplace to the array of every block."""
        for sector, array in self.get_sector_block_pairs():
            self.set_block(sector, fn(array))

    def item(self):
        """Convert the block array to a scalar if it is a scalar block array."""
        (array,) = self.get_all_blocks()
        return array.item()

    def _binary_blockwise_op_abelian(
        self, other, fn, missing=None, inplace=False
    ):
        """Apply a binary blockwise operation to two block arrays, which must
        have exactly the same sectors/keys, depending on `missing`.

        Parameters
        ----------
        fn : callable
            Function to apply to the blocks of the arrays, with signature
            ``fn(x_block, y_block) -> result_block``.
        x : BlockCommon
            First block array.
        y : BlockCommon
            Second block array.
        missing : str, optional
            How to handle missing blocks. Can be "outer", "inner" or None.
            Default is None which requires 1:1 matching blocks. If missing
            "outer" is specified, blocks present in only one of either array
            are simply kept. If missing "inner" is specified, blocks present
            in only one of either array are simply dropped.
        inplace : bool, optional
            Whether to modify the first array in place. Default is False.

        Returns
        -------
        BlockCommon
        """
        xy = self if inplace else self.copy()

        xy_blocks = xy.blocks
        other_blocks = other.blocks.copy()

        if missing is None:
            # by default require 1:1 matching blocks

            for sector, x_block in xy_blocks.items():
                try:
                    other_block = other_blocks.pop(sector)
                except KeyError:
                    raise ValueError(
                        f"Not all left blocks present in right: {sector}"
                    )
                xy_blocks[sector] = fn(x_block, other_block)

            if other_blocks:
                raise ValueError(
                    f"Not all right blocks present in left: {other_blocks}."
                )

        elif missing == "outer":
            # blocks present in only one of either array are simply kept

            for sector, x_block in xy_blocks.items():
                if sector in other_blocks:
                    # both blocks present -> combine with fn
                    other_block = other_blocks.pop(sector)
                    xy_blocks[sector] = fn(x_block, other_block)
                else:
                    # only left present
                    xy_blocks[sector] = x_block

            # add remaining blocks only present in right
            xy_blocks.update(other_blocks)

        elif missing == "inner":
            # blocks present in only one of either array are simply dropped

            for sector, x_block in xy_blocks.items():
                if sector in other_blocks:
                    other_block = other_blocks.pop(sector)
                    xy_blocks[sector] = fn(x_block, other_block)

        return xy

    def _do_reduction(self, fn):
        """Perform an (associative) reduction operation on blocks of the array."""
        if isinstance(fn, str):
            fn = ar.get_lib_fn(self.backend, fn)
            _stack = ar.get_lib_fn(self.backend, "stack")
        block_results = tuple(map(fn, self.get_all_blocks()))
        return fn(_stack(block_results))

    def norm(self):
        """Get the frobenius norm of the block array."""
        backend = self.backend
        _sum = ar.get_lib_fn(backend, "sum")
        _abs = ar.get_lib_fn(backend, "abs")
        return (
            functools.reduce(
                operator.add,
                (_sum(_abs(x) ** 2) for x in self.get_all_blocks()),
            )
            ** 0.5
        )

    def _allclose_blockcommon(self, other, **allclose_opts):
        _allclose = ar.get_lib_fn(self.backend, "allclose")

        # all shared blocks must be close
        shared = self.blocks.keys() & other.blocks.keys()
        for sector in shared:
            if not _allclose(
                self.get_block(sector),
                other.get_block(sector),
                **allclose_opts,
            ):
                return False

        # all missing blocks must be zero
        left = self.blocks.keys() - other.blocks.keys()
        right = other.blocks.keys() - self.blocks.keys()
        for sector in left:
            if not _allclose(self.get_block(sector), 0.0, **allclose_opts):
                return False
        for sector in right:
            if not _allclose(other.get_block(sector), 0.0, **allclose_opts):
                return False

        return True

    @lazyabstractmethod
    def allclose(self, other, **allclose_opts):
        pass

    def _test_allclose_blockcommon(self, other, **allclose_opts):
        """Like `allclose` but raises an AssertionError with details if not
        close."""
        _allclose = ar.get_lib_fn(self.backend, "allclose")

        # all shared blocks must be close
        shared = self.blocks.keys() & other.blocks.keys()
        for sector in shared:
            if not _allclose(
                self.get_block(sector),
                other.get_block(sector),
                **allclose_opts,
            ):
                raise AssertionError(
                    f"Block arrays are not allclose for sector {sector}:\n"
                    f" Left  block:\n{self.get_block(sector)}\n"
                    f" Right block:\n{other.get_block(sector)}\n"
                )

        # all missing blocks must be zero
        left = self.blocks.keys() - other.blocks.keys()
        right = other.blocks.keys() - self.blocks.keys()
        for sector in left:
            if not _allclose(self.get_block(sector), 0.0, **allclose_opts):
                raise AssertionError(
                    f"Left block only is not zero for {sector}:\n"
                    f" Left  block:\n{self.get_block(sector)}\n"
                    f" Right block:\n<missing>\n"
                )
        for sector in right:
            if not _allclose(other.get_block(sector), 0.0, **allclose_opts):
                raise AssertionError(
                    f"Right block only is not zero for {sector}:\n"
                    f" Left  block:\n<missing>\n"
                    f" Right block:\n{other.get_block(sector)}\n"
                )

        return True

    @lazyabstractmethod
    def test_allclose(self, other, **allclose_opts):
        pass



"""Flat backend block vector class."""

import numbers

import autoray as ar

from ..common import SymmrayCommon
from ..utils import DEBUG
from ..vector_common import VectorCommon
from .flat_data_common import FlatCommon


class FlatVector(FlatCommon, VectorCommon, SymmrayCommon):
    """Class for storing block vectors with flat storage, e.g. for the
    singular- or eigen- values of a matrix.

    This is equivalent to the diagonal of a zero charge abelian matrix.

    Parameters
    ----------
    sectors : array_like
        The vector of charges, with shape (num_blocks,).
    blocks : array_like
        The stack of vectors, with shape (num_blocks, charge_subsize).
    """

    __slots__ = ("_blocks", "_sectors", "backend")

    def __init__(self, sectors, blocks):
        self._init_flatcommon(sectors, blocks)
        if DEBUG:
            self.check()

    @classmethod
    def from_blocks(cls, blocks):
        """Create a flat vector from a dictionary of blocks.

        Parameters
        ----------
        blocks : dict
            Dictionary mapping sector (charge) to block (array).
        """
        new_blocks = ar.do("stack", tuple(blocks.values()))
        sectors = ar.do("array", tuple(blocks.keys()), like=new_blocks)
        return cls(sectors, new_blocks)

    @classmethod
    def from_blockvector(cls, blockvector):
        """Create a flat backend vector from a sparse backend BlockVector.

        Parameters
        ----------
        blockvector : BlockVector
            The BlockVector to convert.
        """
        return cls.from_blocks(blockvector.blocks)

    def to_blockvector(self):
        """Convert to a sparse backend BlockVector."""
        from ..sparse.sparse_vector import BlockVector

        return BlockVector(
            {k.item(): b for k, b in zip(self.sectors, self.blocks)}
        )

    @classmethod
    def from_fill_fn(cls, fill_fn, sectors, charge_size):
        """Create a flat vector by filling blocks using a function that takes
        a shape and returns an array.

        Parameters
        ----------
        fill_fn : callable
            Function that takes a shape tuple and returns an array of that
            shape.
        sectors : array_like or int or str
            The vector of charges, with shape (num_blocks,). If an integer
            is given, it is assumed to be Z{N} symmetry and the sectors are
            set to range(N). If a string is given, it is assumed to be of the
            form "ZN" for some integer N.
        charge_size : int
            The size of each charge block.
        like : str or array_like, optional

        """
        if isinstance(sectors, str):
            # assume {ZN} symmetry -> turn into int
            import re

            m = re.match(r"Z(\d+)", sectors)
            if m is None:
                raise ValueError(f"Invalid symmetry string: {sectors}")
            sectors = int(m[1])

        if isinstance(sectors, numbers.Integral):
            # integer specifying Z{N} symmetry
            n = sectors
            create_sectors = True
        else:
            n = len(sectors)
            create_sectors = False

        block_shape = (n, charge_size)
        blocks = fill_fn(block_shape)

        if create_sectors:
            sectors = ar.do("arange", n, like=blocks)

        return cls(sectors, blocks)

    @classmethod
    def rand(
        cls, sectors, charge_size, dist="normal", seed=None, like="numpy"
    ):
        rng = ar.do("random.default_rng", seed, like=like)

        if dist == "normal":

            def fill_fn(shape):
                return rng.normal(size=shape)

        elif dist == "uniform":

            def fill_fn(shape):
                return rng.uniform(size=shape)

        else:
            raise ValueError(f"Invalid distribution: {dist}")

        return cls.from_fill_fn(fill_fn, sectors, charge_size)

    @property
    def size(self):
        """The total size of all elements in the vector."""
        db, dv = self._get_shape_blocks_full()
        return db * dv

    @property
    def shape(self) -> tuple[int, ...]:
        """Get the effective shape of the vector."""
        return (self.size,)

    def copy(self, deep=False) -> "FlatVector":
        return self._copy_flatcommon(deep=deep)

    def copy_with(self, sectors=None, blocks=None) -> "FlatVector":
        """Create a copy of the vector with some attributes replaced. Note that
        checks are not performed on the new properties, this is intended for
        internal use.
        """
        new = self._copy_with_flatcommon(sectors=sectors, blocks=blocks)
        if DEBUG:
            new.check()
        return new

    def modify(self, sectors=None, blocks=None):
        """Modify the vector in place with some attributes replaced. Note that
        checks are not performed on the new properties, this is intended for
        internal use.
        """
        self._modify_flatcommon(sectors=sectors, blocks=blocks)
        if DEBUG:
            self.check()
        return self

    def check(self):
        assert ar.do("ndim", self._blocks, like=self.backend) == 2

    def sort_stack(self, inplace=False):
        """Sort the sectors and associated blocks in ascending order of
        charge.
        """
        k = ar.do("argsort", self._sectors, like=self.backend)
        return self._modify_or_copy(
            sectors=self._sectors[k],
            blocks=self._blocks[k],
            inplace=inplace,
        )

    def _binary_blockwise_op(self, other, fn, missing=None, inplace=False):
        return self._binary_blockwise_op_abelian(
            other, fn, missing=missing, inplace=inplace
        )

    def to_dense(self):
        return ar.do("reshape", self._blocks, (-1,), like=self.backend)

    def __repr__(self):
        return "".join(
            [
                f"{self.__class__.__name__}(",
                f"total_size={self.size}, ",
                f"num_blocks={self.num_blocks})",
            ]
        )



import operator

from .common import SymmrayCommon
from .utils import DEBUG


class VectorCommon:
    def __add__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            # can directly add matching vectors
            return self._binary_blockwise_op(
                other, fn=operator.add, missing="outer", inplace=inplace
            )

        if isinstance(other, SymmrayCommon):
            # cannot add to symmray objects:
            # sparsity not preserved or incompatible backend
            # XXX: possibly support 0-dimensional symmray objects?
            return NotImplemented

        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Addition {self} + {other} not supported.")

        # assume scalar
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x + other)
        return new

    def __iadd__(self, other):
        return self.__add__(other, inplace=True)

    def __radd__(self, other):
        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Addition {other} + {self} not supported.")
        # assume scalar
        new = self.copy()
        new.apply_to_arrays(lambda x: other + x)
        return new

    def __sub__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            # can directly subtract matching vectors
            return self._binary_blockwise_op(
                other, fn=operator.sub, inplace=inplace
            )

        if isinstance(other, SymmrayCommon):
            # block structure not preserved
            # XXX: possibly support 0-dimensional symmray objects?
            return NotImplemented

        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Subtraction {self} - {other} not supported.")

        # assume scalar
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x - other)
        return new

    def __isub__(self, other):
        return self.__sub__(other, inplace=True)

    def __rsub__(self, other):
        # assume scalar
        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Subtraction {other} - {self} not supported.")
        new = self.copy()
        new.apply_to_arrays(lambda x: other - x)
        return new

    def __truediv__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            # can directly divide matching vectors
            return self._binary_blockwise_op(
                other, fn=operator.truediv, inplace=inplace
            )

        if isinstance(other, SymmrayCommon):
            # deviding by implicit zeros not defined
            # XXX: possibly support 0-dimensional symmray objects?
            return NotImplemented

        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Division {self} / {other} not supported.")

        # assume scalar
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x / other)
        return new

    def __itruediv__(self, other):
        return self.__truediv__(other, inplace=True)

    def __rtruediv__(self, other):
        # assume scalar
        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Division {other} / {self} not supported.")
        new = self.copy()
        new.apply_to_arrays(lambda x: other / x)
        return new

    def __pow__(self, other, inplace=False):
        if isinstance(other, self.__class__):
            return self._binary_blockwise_op(
                other, fn=operator.pow, inplace=inplace
            )

        if isinstance(other, SymmrayCommon):
            # exponentiation with symmray objects not defined
            # XXX: possibly support 0-dimensional symmray objects?
            return NotImplemented

        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Exponential {self} ** {other} not supported.")

        # assume scalar
        new = self if inplace else self.copy()
        new.apply_to_arrays(lambda x: x**other)
        return new

    def __ipow__(self, other):
        return self.__pow__(other, inplace=True)

    def __rpow__(self, other):
        if DEBUG and getattr(other, "ndim", 0) != 0:
            raise ValueError(f"Exponential {other} ** {self} not supported.")
        # assume scalar
        new = self.copy()
        new.apply_to_arrays(lambda x: other**x)
        return new
